@@ -13,7 +13,7 @@
  * bb_getopt_ulflags();
  */
 
-/* Design notes: There is no spec for this.  Remind me to write one.
+/* Design notes: There is no spec for mount.  Remind me to write one.
 
    mount_main() calls singlemount() which calls mount_it_now().
 
@@ -22,18 +22,14 @@
    mount_it_now() does the actual mount.
 */
 
-#include <limits.h>
-#include <stdlib.h>
+#include "busybox.h"
 #include <unistd.h>
 #include <errno.h>
 #include <string.h>
-#include <stdio.h>
 #include <mntent.h>
 #include <ctype.h>
-#include <sys/mount.h>
 #include <fcntl.h>		// for CONFIG_FEATURE_MOUNT_LOOP
 #include <sys/ioctl.h>  // for CONFIG_FEATURE_MOUNT_LOOP
-#include "busybox.h"
 
 // These two aren't always defined in old headers
 #ifndef MS_BIND
@@ -169,7 +165,7 @@ static llist_t *get_block_backed_filesystems(void)
 			if(*fs=='#' || *fs=='*') continue;
 			if(!*fs) continue;
 
-			list=llist_add_to_end(list,bb_xstrdup(fs));
+			llist_add_to_end(&list,bb_xstrdup(fs));
 		}
 		if (ENABLE_FEATURE_CLEAN_UP) fclose(f);
 	}
@@ -182,8 +178,10 @@ llist_t *fslist = 0;
 #if ENABLE_FEATURE_CLEAN_UP
 static void delete_block_backed_filesystems(void)
 {
-	llist_free(fslist);
+	llist_free(fslist, free);
 }
+#else
+void delete_block_backed_filesystems(void);
 #endif
 
 #if ENABLE_FEATURE_MTAB_SUPPORT
@@ -196,12 +194,9 @@ static int fakeIt;
 
 // Perform actual mount of specific filesystem at specific location.
 
-static int mount_it_now(struct mntent *mp, int vfsflags)
+static int mount_it_now(struct mntent *mp, int vfsflags, char *filteropts)
 {
 	int rc;
-	char *filteropts = 0;
-
-	parse_mount_options(mp->mnt_opts, &filteropts);
 
 	if (fakeIt) { return 0; }
 
@@ -217,8 +212,6 @@ static int mount_it_now(struct mntent *mp, int vfsflags)
 		vfsflags |= MS_RDONLY;
 	}
 
-    free(filteropts);
-
 	// Abort entirely if permission denied.
 
 	if (rc && errno == EPERM)
@@ -263,14 +256,14 @@ static int mount_it_now(struct mntent *mp, int vfsflags)
 // Mount one directory.  Handles NFS, loopback, autobind, and filesystem type
 // detection.  Returns 0 for success, nonzero for failure.
 
-static int singlemount(struct mntent *mp)
+static int singlemount(struct mntent *mp, int ignore_busy)
 {
-	int rc = 1, vfsflags;
-	char *loopFile = 0;
+	int rc = -1, vfsflags;
+	char *loopFile = 0, *filteropts = 0;
 	llist_t *fl = 0;
 	struct stat st;
 
-	vfsflags = parse_mount_options(mp->mnt_opts, 0);
+	vfsflags = parse_mount_options(mp->mnt_opts, &filteropts);
 
 	// Treat fstype "auto" as unspecified.
 
@@ -282,23 +275,24 @@ static int singlemount(struct mntent *mp)
 		(!mp->mnt_type || !strcmp(mp->mnt_type,"nfs")) &&
 		strchr(mp->mnt_fsname, ':') != NULL)
 	{
-		char *options=0;
-		parse_mount_options(mp->mnt_opts, &options);
-		if (nfsmount(mp->mnt_fsname, mp->mnt_dir, &vfsflags, &options, 1)) {
+		if (nfsmount(mp->mnt_fsname, mp->mnt_dir, &vfsflags, &filteropts, 1)) {
 			bb_perror_msg("nfsmount failed");
-			return 1;
+			goto report_error;
+		} else {
+			// Strangely enough, nfsmount() doesn't actually mount() anything.
+			mp->mnt_type = "nfs";
+			rc = mount_it_now(mp, vfsflags, filteropts);
+			if (ENABLE_FEATURE_CLEAN_UP) free(filteropts);
+			
+			goto report_error;
 		}
-
-		// Strangely enough, nfsmount() doesn't actually mount() anything.
-
-		else return mount_it_now(mp, vfsflags);
 	}
 
-	// Look at the file.  (Not found isn't a failure for remount.)
+	// Look at the file.  (Not found isn't a failure for remount, or for
+	// a synthetic filesystem like proc or sysfs.)
 
-	if (lstat(mp->mnt_fsname, &st));
-
-	if (!(vfsflags & (MS_REMOUNT | MS_BIND | MS_MOVE))) {
+	if (stat(mp->mnt_fsname, &st));
+	else if (!(vfsflags & (MS_REMOUNT | MS_BIND | MS_MOVE))) {
 		// Do we need to allocate a loopback device for it?
 
 		if (ENABLE_FEATURE_MOUNT_LOOP && S_ISREG(st.st_mode)) {
@@ -324,7 +318,7 @@ static int singlemount(struct mntent *mp)
 	 * to the actual mount. */
 
 	if (mp->mnt_type || (vfsflags & (MS_REMOUNT | MS_BIND | MS_MOVE)))
-		rc = mount_it_now(mp, vfsflags);
+		rc = mount_it_now(mp, vfsflags, filteropts);
 
 	// Loop through filesystem types until mount succeeds or we run out
 
@@ -336,33 +330,38 @@ static int singlemount(struct mntent *mp)
 
 		if (!fslist) {
 			fslist = get_block_backed_filesystems();
-#if ENABLE_FEATURE_CLEAN_UP
 			if (ENABLE_FEATURE_CLEAN_UP && fslist)
 				atexit(delete_block_backed_filesystems);
-#endif
 		}
 
 		for (fl = fslist; fl; fl = fl->link) {
 			mp->mnt_type = fl->data;
 
-			if (!(rc = mount_it_now(mp,vfsflags))) break;
+			if (!(rc = mount_it_now(mp,vfsflags, filteropts))) break;
 
 			mp->mnt_type = 0;
 		}
 	}
 
-	// Mount failed.  Clean up
-	if (rc && loopFile) {
+	if (ENABLE_FEATURE_CLEAN_UP) free(filteropts);
+
+	// If mount failed, clean up loop file (if any).
+
+	if (ENABLE_FEATURE_MOUNT_LOOP && rc && loopFile) {
 		del_loop(mp->mnt_fsname);
-		if(ENABLE_FEATURE_CLEAN_UP) {
+		if (ENABLE_FEATURE_CLEAN_UP) {
 			free(loopFile);
 			free(mp->mnt_fsname);
 		}
 	}
+report_error:
+	if (rc && errno == EBUSY && ignore_busy) rc = 0;
+	if (rc < 0)
+		bb_perror_msg("Mounting %s on %s failed", mp->mnt_fsname, mp->mnt_dir);
+
 	return rc;
 }
 
-
 // Parse options, if necessary parse fstab/mtab, and call singlemount for
 // each directory to be mounted.
 
@@ -370,7 +369,7 @@ int mount_main(int argc, char **argv)
 {
 	char *cmdopts = bb_xstrdup(""), *fstabname, *fstype=0, *storage_path=0;
 	FILE *fstab;
-	int i, opt, all = FALSE, rc = 1;
+	int i, opt, all = FALSE, rc = 0;
 	struct mntent mtpair[2], *mtcur = mtpair;
 
 	/* parse long options, like --bind and --move.  Note that -o option
@@ -452,7 +451,7 @@ int mount_main(int argc, char **argv)
 		mtpair->mnt_dir = argv[optind+1];
 		mtpair->mnt_type = fstype;
 		mtpair->mnt_opts = cmdopts;
-		rc = singlemount(mtpair);
+		rc = singlemount(mtpair, 0);
 		goto clean_up;
 	}
 
@@ -477,8 +476,9 @@ int mount_main(int argc, char **argv)
 
 		// Get next fstab entry
 
-		if (!getmntent_r(fstab, mtcur, bb_common_bufsiz1,
-					sizeof(bb_common_bufsiz1)))
+		if (!getmntent_r(fstab, mtcur, bb_common_bufsiz1
+					+ (mtcur==mtpair ? sizeof(bb_common_bufsiz1)/2 : 0),
+				sizeof(bb_common_bufsiz1)/2))
 		{
 			// Were we looking for something specific?
 
@@ -495,10 +495,10 @@ int mount_main(int argc, char **argv)
 				mtcur = mtnext;
 				mtcur->mnt_opts=bb_xstrdup(mtcur->mnt_opts);
 				append_mount_options(&(mtcur->mnt_opts),cmdopts);
-				rc = singlemount(mtcur);
+				rc = singlemount(mtcur, 0);
 				free(mtcur->mnt_opts);
 			}
-			break;
+			goto clean_up;
 		}
 
 		/* If we're trying to mount something specific and this isn't it,
@@ -533,11 +533,9 @@ int mount_main(int argc, char **argv)
 
 			// Mount this thing.
 
-			rc = singlemount(mtcur);
-			if (rc) {
-				// Don't whine about already mounted fs when mounting all.
-				if (errno == EBUSY) rc = 0;
-				else break;
+			if (singlemount(mtcur, 1)) {
+				/* Count number of failed mounts */
+				rc++;
 			}
 		}
 	}
@@ -548,12 +546,7 @@ clean_up:
 	if (ENABLE_FEATURE_CLEAN_UP) {
 		free(storage_path);
 		free(cmdopts);
-		free(fstype);
 	}
 
-	if(rc)
-		bb_perror_msg("Mounting %s on %s failed",
-				mtcur->mnt_fsname, mtcur->mnt_dir);
-
 	return rc;
 }
